/****************************** Module Header *******************************
*
* Module Name: undo.e
*
* Copyright (c) Netlabs EPM Distribution Project 2002
*
* $Id: undo.e 3568 2019-04-23 12:12:39Z aschn $
*
* ===========================================================================
*
* This file is part of the Netlabs EPM Distribution package and is free
* software.  You can redistribute it and/or modify it under the terms of the
* GNU General Public License as published by the Free Software
* Foundation, in version 2 as it comes in the "COPYING" file of the
* Netlabs EPM Distribution.  This library is distributed in the hope that it
* will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
****************************************************************************/

; ---------------------------------------------------------------------------
; Create a new state, if not done before.
defproc AddUndoStateIfNew
   --universal curkey

   --PrevStateRange = QueryUndoState()

   undoaction 1, JunkState

   --NewStateRange = QueryUndoState()
   --if PrevStateRange <> NewStateRange then
      --dprintf( 'AddUndoStateIfNew( 'arg( 1)'): PrevStateRange = 'PrevStateRange', NewStateRange = 'NewStateRange', curkey = 'curkey)
   --endif

   return

; ---------------------------------------------------------------------------
; This won't create a new undo state.
; StateRange consists of two integers, seaparated by a space:
; StateRange = OldestState NewestState
defproc QueryUndoState
   undoaction 6, StateRange

   -- During LoadAttributes, the oldest undo state may be set to limit
   -- the oldest undo state to 2 instead of 1.
   -- The undo state itself can't be avoided, but the StateRange query
   -- here can be changed. The undoaction statement remains unchanged from
   -- that, because it's internally defined.
   getfileid fid
   SavedOldestState = GetAVar( fid'.undooldeststate')
   if IsNum( SavedOldestState) then
      parse value StateRange with OldestState NewestState
      StateRange = Max( SavedOldestState, OldestState) NewestState
   endif

   return StateRange

; ---------------------------------------------------------------------------
; Restore undo state to a specified state. This creates an additional undo
; state if undo recording is enabled. Therefore it is disabled during undo
; processing and reenabled by a next key.
defproc GotoUndoState
   State = arg( 1)
   undoaction 7, State
   return

; ---------------------------------------------------------------------------
; Disable undo recording at a specific event.
defproc DisableUndoEvent
   Event = arg( 1)
   if Event = 'K' then
      Event = 1
   elseif Event = 'C' then
      Event = 2
   elseif Event = 'L' then
      Event = 3
   endif
   undoaction 4, Event
   return

; ---------------------------------------------------------------------------
; Enable undo recording at a specific event.
defproc EnableUndoEvent
   Event = arg( 1)
   if Event = 'K' then
      Event = 1
   elseif Event = 'C' then
      Event = 2
   elseif Event = 'L' then
      Event = 3
   endif
   undoaction 5, Event
   return

; ---------------------------------------------------------------------------
; Enable or disable undo state creation at several events.
;    K = upon starting each keystroke
;    C = upon starting each command
;    L = when moving the cursor from a modified line.
; Syntax: ConfigUndoRec( Events)
;    Events = combination of event letters
;    Example: ConfigUndoRec( 'LKC')
; All undo events are disabled, if '' is set as Events.
;
; The proc name 'UndoRec' apparently is internally used and leads to
; 'unknown procedure' when being called by DisableUndoRe or
; EnableUndoRec. Therefore the proc was named 'ConfigUndoRec'.
defproc ConfigUndoRec
   Events = upcase( arg( 1))

   Event = 0  -- upon starting each keystroke
   if pos( 'K', Events) > 0 then
      call EnableUndoEvent( Event)
   else
      call DisableUndoEvent( Event)
   endif

   Event = 1  -- upon starting each command
   if pos( 'C', Events) > 0 then
      call EnableUndoEvent( Event)
   else
      call DisableUndoEvent( Event)
   endif

   Event = 2  -- when moving the cursor from a modified line
   if pos( 'L', Events) > 0 then
      call EnableUndoEvent( Event)
   else
      call DisableUndoEvent( Event)
   endif

   -- Creating undo states for these events is processed internally.
   -- It can be controlled by Enable/DisableUndoEvent or undoaction.
   -- Normally the above events are disabled. That's because no undo
   -- positions are saved for them.
   -- Additionally these event letters were added:
   --    A = on any key if key and line differ from previous ones
   --    N = newline or enter key
   --    S = space key
   --    O = other calls of NewUndoRec or CreateUndoState
   -- See CreateUndoState.

   -- Save the current event string
   SetAVar( 'undorec', Events)

   return

; ---------------------------------------------------------------------------
; Temporary disable undo.
defc DisableUndoRec
   UndoRec = GetAVar( 'undorec')
   SavedUndoRec = GetAVar( 'savedundorec')

   if UndoRec <> '' then
      SetAVar( 'savedundorec', UndoRec)
      call ConfigUndoRec( '')
   endif

; ---------------------------------------------------------------------------
; Enable undo back.
defc EnableUndoRec
   UndoRec = GetAVar( 'undorec')
   SavedUndoRec = GetAVar( 'savedundorec')

   if UndoRec <> SavedUndoRec & SavedUndoRec <> '' then
      SetAVar( 'savedundorec', '')
      call ConfigUndoRec( SavedUndoRec)
   endif

; ---------------------------------------------------------------------------
; Command for setting undo behavior from REXX and C
defc UndoRec, ConfigUndoRec
   parse arg Events
   call ConfigUndoRec( Events)

; ---------------------------------------------------------------------------
; Disable undo recording for all events.
; Undo states were created only during execution of a command that calls
; NextCmdAltersText(), defined in keys.e.
; Enabling any of the predefined events 1 (K), 2 (C)) or 3 (L) would create
; undo states with missing undo positions.
definit
   --    K = upon starting each keystroke
   --    C = upon starting each command
   --    L = when moving the cursor from a modified line
   --    A = on any key if key and line differ from previous ones
   --    N = newline or enter key
   --    S = space key
   --    O = other calls of NewUndoRec or CreateUndoState
   --call ConfigUndoRec( 'LO')
   call ConfigUndoRec( 'ANO')

; ---------------------------------------------------------------------------
; Create a new undo record, if the current state is not already checkpointed.
; This is also called by NextCmdAltersText and only for non-repeated keys.
; NextCmdAltersText is called in each text-altering cmd def. It is called
; before a key cmd alters text.
defproc NewUndoRec
   universal curkey

   -- Keep undo recording disabled for Undo1, Redo1 and UndoDlg keys.
   parse value( curkey) with Key''\1''Cmd
   UndoCmds = 'undo1 redo1 undodlg'
   UndoKeys = ''
   do w = 1 to words( UndoCmds)
      UndoCmd = lowcase( word( UndoCmds, w))
      UndoKey = strip( GetAVar( 'keycmd.'UndoCmd))
      if UndoKeys = '' then
         UndoKeys = UndoKey
      else
         UndoKeys = UndoKeys UndoKey
      endif
   enddo

   if wordpos( Key, UndoKeys) = 0 then  -- current key is not undo or redo
      -- Other keys reenable undo recording.
      --if GetAVar( 'savedundorec') <> '' then
      --   dprintf( '* Enable undo recording by 'Key)
      --endif
      'EnableUndoRec'
   endif

   call CreateUndoState( arg( 1))

   return

; ---------------------------------------------------------------------------
; Command for creating undo states from REXX and C
defc NewUndoRec
   call NewUndoRec( arg( 1))

; ---------------------------------------------------------------------------
; Helper proc.
defproc SetUndoStatePos( fid, NewState)
   call psave_pos( savepos)
   --call psave_mark( savemark)  -- can't use psave_mark here, it unmarks text
   getmark savefirstline, savelastline, savefirstcol, savelastcol, savemkfileid
   if savemkfileid = fid then
      savemt = marktype()
      savemark = savefirstline savelastline savefirstcol savelastcol savemkfileid savemt
   else
      savemark = ''
   endif
   --dprintf( 'SetUndoStatePos: 'fid'.'NewState'.undostatepos = 'savepos'|'savemark)
   call SetAvar( fid'.'NewState'.undostatepos', savepos'|'savemark)
   return

; ---------------------------------------------------------------------------
; This is called by defload to init the array var.
; Otherwise the undo state '1' won't have a pos var.
defproc InitUndo
   fid = arg(1)
   if fid = '' then
      getfileid fid
   endif
   if fid.modify = 0 then
      if QueryUndoState() = '1 1' then
         NewState = 1
         if GetAvar( fid'.'NewState'.undostatepos') = '' then
            --dprintf( 'InitUndo')
            call SetUndoStatePos( fid, NewState)
         endif
      endif
   endif

; ---------------------------------------------------------------------------
; This checkpoints a new undo state, if not done before.
; This is intended to be called by code or keys that should trigger the
; creation of a new undo state.
; Calling this after undoaction 7 is not required and not recommended: It
; would always create an additional state.
; This is not performed for unvisible (temporary) files.
; This saves cursor pos, scroll pos, mark pos and mark type
; per undo state and file.
defproc CreateUndoState
   do once = 1 to 1
      if not .visible then
         leave
      endif

      -- Filter out disabled events
      EnabledEvents = upcase( GetAVar( 'undorec'))
      fEnableKeyLineChange = pos( 'A', EnabledEvents)
      fEnableNewline       = pos( 'N', EnabledEvents)
      fEnableSpace         = pos( 'S', EnabledEvents)
      fEnableOther         = pos( 'O', EnabledEvents)
      parse value arg( 1) with 1 Event 2 .
      Event = strip( upcase( Event))
      if not wordpos( Event, 'A N S') then
         Event = 'O'
      endif
      if Event = 'A' & not fEnableKeyLineChange then    -- A = on any key if key and line have changed
         leave
      endif
      if Event = 'N' & not fEnableNewLine then          -- N = on Newline and Enter
         leave
      endif
      if Event = 'S' & not fEnableSpace then            -- S = on Space
         leave
      endif
      if Event = 'O' & not fEnableOther then            -- O = on other commands that call NewUndoRec or CreateUndoState
         leave
      endif

      -- Create a new undo state if required
      AddUndoStateIfNew( arg( 1))

      -- Query new state
      StateRange = QueryUndoState()
      parse value StateRange with OldestState NewestState
      NewState = NewestState
      --dprintf( 'CreateUndoState: NewState = 'NewState', StateRange = 'StateRange)

      getfileid fid

      -- Check if pos. for NewState was already saved. That may happen when
      -- NewState wasn't changed.
      next = GetAvar( fid'.'NewState'.undostatepos')
      parse value next with savepos'|'savemark
      parse value savepos with svline svcol svcx svcy
      if svcy <> '' then
         --dprintf( 'CreateUndoState: NewState = 'NewState', pos exists')
         leave
      endif

      -- Save cursor, scroll and mark pos. for the new state
      --dprintf( 'CreateUndoState: NewState = 'NewState', SetUndoStatePos( 'fid', 'NewState')')
      call SetUndoStatePos( fid, NewState)
   enddo
   return

; ---------------------------------------------------------------------------
; This performs an undo action to a state submitted as arg( 1).
; For repeated undoaction 7 calls, the first one always creates a new undo
; state.
; This restores cursor pos, scroll pos, mark pos and mark type
; per undo state and file.
; The Undo dialog won't show the cursor on scrolling through the undo states.
defproc SelectUndoState
   universal current_undo_state

   SelectedState = arg( 1)

   parse value current_undo_state with PresentState OldestState NewestState
   SavedPresentState = PresentState
   getfileid fid

   -- Restore undo state or return
   if not IsNum( SelectedState) then
      --dprintf( 'SelectUndoState: Not IsNum, current_undo_state = 'current_undo_state', return')
      return
   elseif SelectedState = PresentState then
      --dprintf( 'SelectUndoState: Selected = Present, current_undo_state = 'current_undo_state', return')
      return
   else
      --dprintf( 'SelectUndoState: current_undo_state = 'current_undo_state', GotoUndoState( 'SelectedState')')
      -- Restore undo state SelectedState
      call GotoUndoState( SelectedState)
      PresentState = SelectedState
      current_undo_state = PresentState OldestState NewestState
      --dprintf( 'SelectUndoState: Set current_undo_state = 'current_undo_state)
   endif

   -- Restore pos. and highlight cursor
   do once = 1 to 1
      next = GetAvar( fid'.'SelectedState'.undostatepos')
      --dprintf( 'SelectUndoState: 'fid'.'SelectedState'.undostatepos = 'next)
      parse value next with savepos'|'savemark
      parse value savepos with svline svcol svcx svcy

      -- Check parameters
      if svline = '' | svcol = '' then
         -- Don't try to restore pos.
         leave
      endif

      LastUndoType = GetAVar( fid'.lastundotype')  -- Query last type
      if SelectedState >= SavedPresentState then  -- Undo
         UndoType = 'R'      -- Redo
      else
         if LastUndoType = '' | LastUndoType = 'R' then
            UndoType = 'UT'  -- Undo, toggle direction
         else
            UndoType = 'U'   -- Undo
         endif
      endif
      call SetAVar( fid'.lastundotype', UndoType)  -- Save current as last type
      --dprintf( 'SelectUndoState: UndoType = 'UndoType', LastUndoType = 'LastUndoType)

      -- If (Undo and (last type was Redo or unset)) or undo state 1:
      -- Keep scroll position and set only line and column.
      -- For other Undo and Redo actions: Use saved scroll position of
      -- selected state (savepos above).
      if UndoType = 'UT' | SelectedState = 1 then
         --dprintf( 'SelectUndoState: -> 1')
         .col   = svcol                   -- set .col
         .lineg = svline                  -- set .line without scrolling
      else
         --dprintf( 'SelectUndoState: -> 2 prestore_pos_min_scroll( 'savepos')')
         call prestore_pos_min_scroll( savepos)
      endif

      'HighlightCursor'
   enddo

   -- Restore mark
   do once = 1 to 1
      if savemark = '' then
         -- Don't unmark other files
         if FileIsMarked() then
            unmark
         endif
      else
         -- Check parameters
         parse value savemark with savefirstline savelastline savefirstcol savelastcol savemkfileid savemt
         if savefirstline = '' | savelastline = '' | savefirstcol = '' | savelastcol = '' |
            savemkfileid = '' | savemt = '' then
            leave
         endif

         call prestore_mark( savemark)
      endif
   enddo

   -- Invalidate array of multi-line comment positions, used by pAssist()
   if IsADefProc( 'PruneAssistArray') then
      call PruneAssistArray()
   endif
   'MatchFindOnMove'

   return

; ---------------------------------------------------------------------------
defc UndoDlg
   call windowmessage( 0,  getpminfo(APP_HANDLE),
                       5131,      -- EPM_POPUNDODLG
                       0,
                       0)

; ---------------------------------------------------------------------------
; Called by undo dialog on changing the undo state back and forth and on
; closing the dialog.
; This may create an additional undo state, caused by
; SelectUndoState -> GotoUndoState -> undoaction 7.
defc ProcessUndo
   NewState = arg( 1)
   call SelectUndoState( NewState)

; ---------------------------------------------------------------------------
; This stores the current undo states PresentState OldestState NewestState
; in a universal var. This is used to have fix values, because the real
; value may change with undo or redo processing.
defproc InitUndoStateVar
   universal current_undo_state

   -- Get state information and set the universal
   StateRange = QueryUndoState()
   parse value StateRange with OldestState NewestState

   PresentState = NewestState
   current_undo_state = PresentState OldestState NewestState
   --dprintf( 'InitUndoState: Set current_undo_state = 'current_undo_state)
   --if GetAVar( 'savedundorec') = '' then
   --   dprintf( '* Disable undo recording by InitUndoStateVar')
   --endif

   'DisableUndoRec'

; ---------------------------------------------------------------------------
; Called by undo dialog on opening.
; This doesn't create a new undo state, except for when the current state
; needs to be checkpointed.
; In the original def, the creation of undo states was disabled before the
; windowmessage execution.
defc RenderUndoInfo
   universal current_undo_state

   call CreateUndoState()  -- Do to fix range and for value
   call InitUndoStateVar()

   --Event = 1            -- 1 = upon starting each command
   --undoaction 4, Event  -- 4 = disable

   parse value current_undo_state with PresentState OldestState NewestState

   StateStr = NewestState OldestState\0  -- reversed and zero-terminated
   call windowmessage( 1, arg(1),    -- send message back to dialog
                       32,           -- WM_COMMAND - 0x0020
                       9999,
                       ltoa( offset( StateStr) || selector( StateStr), 10))

; ---------------------------------------------------------------------------
; Called by undo dialog on closing, for both OK and Cancel.
; This doesn't define anything.
; In the original def, the creation of undo states was enabled back.
defc RestoreUndo
   --Event = 1
   --undoaction 5, Event  -- 5 = enable

; ---------------------------------------------------------------------------
; Undo current line.
defc Undo
   undo

   -- Invalidate array of multi-line comment positions, used by pAassist()
   if IsADefProc( 'PruneAssistArray') then
      call PruneAssistArray()
   endif
   'MatchFindOnMove'

; ---------------------------------------------------------------------------
; The following code defines key commands that let you step backwards
; and forwards through the undo states.
; With the switch to accel keys, a new undo state is only created for
; commands different from Undo1 or Redo1.
defc Undo1
   universal current_undo_state
   universal prevkey

   parse value( prevkey) with Key''\1''Cmd
   UndoKeys = strip( GetAVar( 'keycmd.undo1')) strip( GetAVar( 'keycmd.redo1'))
   --dprintf( 'Key = 'Key', UndoKeys = 'UndoKeys)
   if wordpos( Key, UndoKeys) = 0 then  -- last key was not undo or redo

      call CreateUndoState()  -- Do to fix range and for value
      call InitUndoStateVar()
      parse value current_undo_state with PresentState OldestState NewestState

      if NewestState = 1 then
         'SayHint No earlier undo states available'
      else
         PresentState = NewestState - 1
         call SelectUndoState( PresentState)
         'SayHint Initialized at state' PresentState 'of' NewestState
      endif

   else

      parse value current_undo_state with PresentState OldestState NewestState

      if PresentState > OldestState then
         PresentState = PresentState - 1
         call SelectUndoState( PresentState)
         'SayHint Now at state' PresentState 'of' NewestState
      elseif NewestState = 1 then
         'SayHint No earlier undo states available'
      else
         'SayHint Already at state' PresentState 'of' NewestState
      endif

   endif

; ---------------------------------------------------------------------------
defc Redo1
   universal current_undo_state
   universal prevkey

   parse value( prevkey) with Key''\1''Cmd
   UndoKeys = strip( GetAVar( 'keycmd.undo1')) strip( GetAVar( 'keycmd.redo1'))
   --dprintf( 'Key = 'Key', UndoKeys = 'UndoKeys)
   if wordpos( Key, UndoKeys) = 0 then  -- last key was not undo or redo

      call CreateUndoState()  -- Do to fix range and for value
      call InitUndoStateVar()
      parse value current_undo_state with PresentState OldestState NewestState

      'SayHint Initialized at state' PresentState 'of' NewestState

   else

       parse value current_undo_state with PresentState OldestState NewestState

      if PresentState < NewestState then
         PresentState = PresentState + 1
         call SelectUndoState( PresentState)
         'SayHint Now at state' PresentState 'of' NewestState
      else
         'SayHint Already at state' PresentState 'of' NewestState
      endif

   endif

